import csv
import itertools
import logging
import typing
from collections import Counter

from chardet.universaldetector import UniversalDetector

DEFAULT_ENCODING = "utf-8-sig"

logger = logging.getLogger(__name__)

# Register new dialects
csv.register_dialect("excel-semicolon", csv.get_dialect("excel"), delimiter=";")
csv.register_dialect("excel-pipe", csv.get_dialect("excel"), delimiter="|")
csv.register_dialect("unix-tab", csv.get_dialect("unix"), delimiter="\t")
csv.register_dialect("unix-semicolon", csv.get_dialect("unix"), delimiter=";")
csv.register_dialect("unix-pipe", csv.get_dialect("unix"), delimiter="|")


def detect_file_encoding(
    file: typing.Union[typing.IO[bytes], typing.IO[str]], min_confidence: float = 0.8
) -> str:
    orig_file_pos = file.tell()
    file.seek(0)

    encoding = DEFAULT_ENCODING
    detector = UniversalDetector()
    try:
        for line in file.readlines():
            detector.feed(line)
            if detector.done:
                break
        detector.close()
        if detector.result["confidence"] >= min_confidence:
            encoding = detector.result["encoding"]
    except TypeError:
        logger.warn("File provided for encoding detection is not opened in binary mode!")
    finally:
        file.seek(orig_file_pos)

    return encoding


def detect_csv_dialect(file: typing.IO[str], max_lines: typing.Optional[int] = 200) -> str:
    """Detect which csv dialect is most appropriate for counter-like data"""
    histograms = []

    for dialect in csv.list_dialects():
        histogram = csv_line_length_histogram(file, dialect, max_lines)
        histograms.append((histogram, dialect))

    scores = sorted(histograms, key=lambda x: histogram_score(x[0]), reverse=True)
    winner = scores[0][1]

    if name := getattr(file, "name", None):
        logger.debug("The best dialect for the file '%s' is '%s'", name, winner)
    else:
        logger.debug("The best dialect is '%s'", winner)

    return winner


def histogram_score(histogram: typing.Counter[int]) -> int:
    """Calculate the score of the histogram for comprisons"""
    # Basically the more cells the better score
    return sum(k * v for k, v in histogram.items())


def csv_line_length_histogram(
    file: typing.IO[str], dialect: csv.Dialect, max_lines: typing.Optional[int] = 200
) -> typing.Counter[int]:
    orig_file_pos = file.tell()
    file.seek(0)

    try:
        reader = csv.reader(file, dialect=dialect)
        if max_lines:
            reader = itertools.islice(reader, max_lines)
        counter: Counter[int] = Counter()
        for line in reader:
            counter[len(line)] += 1
    finally:
        file.seek(orig_file_pos)

    return counter



import calendar
import datetime
import re
from typing import List, Optional

import dateutil

counter_month_matcher = re.compile(r"^(?P<month>\w{3})-(?P<year>\d{2}(\d{2})?)$")


def parse_counter_month(text: str) -> Optional[datetime.date]:
    """
    Returns a date object extracted from date formatted according to the specification of COUNTER 5
    i. e. Mmm-YYYY. It also allows Mmm-YY to be more flexible
    :param text:
    :return:
    """
    if not text:
        return None
    m = counter_month_matcher.match(text)
    if m:
        year = int(m.group("year"))
        month = m.group("month")
        if year < 50:
            year += 2000
        elif 50 <= year < 100:
            year += 1900
        month_abbrs = list(calendar.month_abbr)
        if month not in month_abbrs:
            return None
        try:
            return datetime.date(year, month_abbrs.index(month), 1)
        except ValueError:
            return None


def parse_date_fuzzy(date_str: str) -> Optional[datetime.date]:
    """
    Uses dateutil.parser to try to parse a date in US format.
    """
    parser_info = dateutil.parser.parserinfo(dayfirst=False)  # US date formats 1/31/2020
    try:
        return dateutil.parser.parse(date_str, parser_info).date()
    except dateutil.parser.ParserError:
        return None


def get_date_range(start_date: datetime.date, end_date: datetime.date) -> List[datetime.date]:
    months = []
    start_date = start_date.replace(day=1)
    while start_date <= end_date:
        months.append(start_date)
        start_date = start_date + datetime.timedelta(days=45)
        start_date = start_date.replace(day=1)

    return months



import typing
from enum import Enum


class ErrorCode(Enum):
    SERVICE_NOT_AVAILABLE = 1000
    SERVICE_BUSY = 1010
    PREPARING_DATA = 1011
    TOO_MANY_REQUESTS = 1020
    INSUFFICIENT_DATA = 1030
    NOT_AUTHORIZED = 2000
    NOT_AUTHORIZED_INSTITUTION = 2010
    INVALID_API_KEY = 2020
    NOT_AUTHORIZED_IP_ADDRESS = 2030
    REPORT_NOT_SUPPORTED = 3000
    REPORT_VERSION_NOT_SUPPORTED = 3010
    INVALID_DATE_ARGS = 3020
    NO_DATA_FOR_DATE_ARGS = 3030
    DATA_NOT_READY_FOR_DATE_ARGS = 3031
    NO_LONGER_AVAILABLE = 3032
    PARTIAL_DATA_RETURNED = 3040
    PARAMETER_NOT_RECOGNIZED = 3050
    INVALID_REPORT_FILTER = 3060
    INCONGRUOUS_REPORT_FILTER = 3061
    INVALID_REPORT_ATTRIBUTE = 3062
    MISSING_REPORT_FILTER = 3070
    MISSING_REPORT_ATTRIBUTE = 3071
    RETURN_DATA_LIMIT_REACHED = 3080

    def __eq__(self, o):
        if isinstance(o, int):
            return self.value == o
        else:
            return super().__eq__(o)


def error_code_to_severity(error_code: typing.Union[None, int, str]) -> str:
    if error_code is None:
        return "Error"
    try:
        error_code = int(error_code)
    except (ValueError, TypeError):
        return "Error"

    if error_code == 0:
        return "Info"
    elif 1 < error_code < 999:
        return "Warning"
    elif error_code == ErrorCode.SERVICE_NOT_AVAILABLE.value:
        return "Error"  # Fatal
    elif error_code == ErrorCode.SERVICE_BUSY.value:
        return "Error"  # Fatal
    elif error_code == ErrorCode.PREPARING_DATA.value:
        return "Warning"
    elif error_code == ErrorCode.TOO_MANY_REQUESTS.value:
        return "Error"  # Fatal
    elif error_code == ErrorCode.INSUFFICIENT_DATA.value:
        return "Error"  # Fatal
    elif error_code == ErrorCode.NOT_AUTHORIZED.value:
        return "Error"
    elif error_code == ErrorCode.NOT_AUTHORIZED_INSTITUTION.value:
        return "Error"
    elif error_code == ErrorCode.INVALID_API_KEY.value:
        return "Error"
    elif error_code == ErrorCode.NOT_AUTHORIZED_IP_ADDRESS.value:
        return "Error"
    elif error_code == ErrorCode.REPORT_NOT_SUPPORTED.value:
        return "Error"
    elif error_code == ErrorCode.REPORT_VERSION_NOT_SUPPORTED.value:
        return "Error"
    elif error_code == ErrorCode.INVALID_DATE_ARGS.value:
        return "Error"
    elif error_code == ErrorCode.NO_DATA_FOR_DATE_ARGS.value:
        return "Error"
    elif error_code == ErrorCode.DATA_NOT_READY_FOR_DATE_ARGS.value:
        return "Error"  # or Warning
    elif error_code == ErrorCode.NO_LONGER_AVAILABLE.value:
        return "Warning"
    elif error_code == ErrorCode.PARTIAL_DATA_RETURNED.value:
        return "Warning"
    elif error_code == ErrorCode.PARAMETER_NOT_RECOGNIZED.value:
        # According to standard this is supposed to be Warning
        # But full data are supposed to be returned once this status
        # occurs. So it is conveted to info.
        return "Info"
    elif error_code == ErrorCode.INVALID_REPORT_FILTER.value:
        return "Warning"  # or Error
    elif error_code == ErrorCode.INCONGRUOUS_REPORT_FILTER.value:
        # According to standard this is supposed to be Warning or Error
        # But full data are supposed to be returned once this status
        # occurs. So it is conveted to info.
        return "Info"
    elif error_code == ErrorCode.INVALID_REPORT_ATTRIBUTE.value:
        # According to standard this is supposed to be Warning or Error
        # But full data are supposed to be returned once this status
        # occurs. So it is conveted to info.
        return "Info"
    elif error_code == ErrorCode.MISSING_REPORT_FILTER.value:
        return "Error"  # or Warning
    elif error_code == ErrorCode.MISSING_REPORT_ATTRIBUTE.value:
        return "Error"  # or Warning
    elif error_code == ErrorCode.RETURN_DATA_LIMIT_REACHED.value:
        return "Warning"
    else:
        return "Error"



import re
import typing
from dataclasses import dataclass, field
from datetime import date

IDS = [
    "DOI",
    "ISBN",
    "Print_ISSN",
    "Online_ISSN",
    "Proprietary",
    "URI",
]


@dataclass(init=False)
class Author:
    name: str
    ISNI: typing.Optional[str]
    ORCID: typing.Optional[str]

    __slots__ = ["name", "ISNI", "ORCID"]

    def __init__(
        self,
        name: str,
        ISNI: typing.Optional[str] = None,
        ORCID: typing.Optional[str] = None,
    ):
        self.name = re.sub(r"\s+", " ", name).strip()
        self.ISNI = ISNI
        self.ORCID = ORCID

    def as_csv(self):
        ids = []
        if self.ISNI is not None:
            ids.append(f"ISNI:{self.ISNI}")
        if self.ORCID is not None:
            ids.append(f"ORCID:{self.ORCID}")

        return f"{self.name} ({';'.join(ids)})"


# TODO since python3.10 there is `slot=True` attr which can really
# make this simpler (no need to override init anymore and define __slots__)
@dataclass(init=False)
class Identifiers:
    """
    Because of dataclass it should have constant memory footprint and
    memory consumption should be lower as well when it contains at last one item
    compared to ordinary dict.

    For comaptiblity reasons it should work as a dict for some operations.
    """

    DOI: typing.Optional[str]
    ISBN: typing.Optional[str]
    Print_ISSN: typing.Optional[str]
    Online_ISSN: typing.Optional[str]
    Proprietary: typing.Optional[str]
    URI: typing.Optional[str]

    __slots__ = IDS

    def __init__(
        self,
        DOI: typing.Optional[str] = None,
        ISBN: typing.Optional[str] = None,
        Print_ISSN: typing.Optional[str] = None,
        Online_ISSN: typing.Optional[str] = None,
        Proprietary: typing.Optional[str] = None,
        URI: typing.Optional[str] = None,
    ):
        # Need to use parent's __setattr__
        # otherwise ImmutableIdentifiers would raise and exception here
        self.DOI = DOI
        self.ISBN = ISBN
        self.Print_ISSN = Print_ISSN
        self.Online_ISSN = Online_ISSN
        self.Proprietary = Proprietary
        self.URI = URI

    def __getitem__(self, item: str):
        if not isinstance(item, str):
            raise TypeError(item)

        if hasattr(self, item):
            res = getattr(self, item)
            if res is None:
                raise KeyError(item)
            else:
                return res
        else:
            raise KeyError(item)

    def __setitem__(self, key: str, value: typing.Optional[typing.Any]):
        if not hasattr(self, key):
            raise KeyError(key)
        setattr(self, key, value)

    def __bool__(self):
        # at least one field set
        return bool(
            self.DOI
            or self.ISBN
            or self.Print_ISSN
            or self.Online_ISSN
            or self.Proprietary
            or self.URI
        )

    def items(self):
        return [(k, getattr(self, k)) for k in self.keys()]

    def keys(self):
        return [e for e in IDS if getattr(self, e, None) is not None]

    def __eq__(self, other):
        if isinstance(other, Identifiers):
            return all(getattr(self, e) == getattr(other, e) for e in IDS)
        elif isinstance(other, dict):
            return other == {e: self[e] for e in IDS if getattr(self, e, None)}


@dataclass
class CounterRecord:
    # The actual vaule
    value: int

    # mandatory, each record should have at least a start date
    start: typing.Optional[date] = None

    # Optional really, if the report is for whole month
    # then the start date (first day of the month) suffice
    end: typing.Optional[date] = None

    # name of the publication
    title: typing.Optional[str] = None

    # ISBN (books), ISSN (periodics), EISSN (same thing only for electronic version of publications)
    title_ids: Identifiers = field(default_factory=Identifiers)

    # name of item
    item: typing.Optional[str] = None

    # ISBN (books), ISSN (periodics), EISSN (same thing only for electronic version of publications)
    item_ids: Identifiers = field(default_factory=Identifiers)

    # When was the item published
    item_publication_date: typing.Optional[str] = None

    # Who it the author of an item
    item_authors: typing.Optional[typing.List[Author]] = None

    # contains more details
    dimension_data: typing.Dict[str, str] = field(default_factory=dict)

    # What actually value of this record means
    metric: typing.Optional[str] = None

    # Entity to which are these data related
    organization: typing.Optional[str] = None

    def __post_init__(self):
        # Try to convert dict in title_ids and item_ids to Identifiers
        if isinstance(self.item_ids, dict):
            self.item_ids = Identifiers(**self.item_ids)
        if isinstance(self.title_ids, dict):
            self.title_ids = Identifiers(**self.title_ids)

    def as_csv(self) -> typing.Tuple[str, ...]:
        def serialize_dict(mapping: typing.Optional[typing.Union[dict, Identifiers]]) -> str:
            if not mapping:
                return ""
            return "|".join(f"{k}:{mapping[k]}" for k in sorted(mapping.keys()))

        def format_date(date_obj: typing.Optional[date]):
            if not date_obj:
                return ""
            return date_obj.strftime("%Y-%m-%d")

        authors = "|".join(e.as_csv() for e in (self.item_authors or []))

        return (
            format_date(self.start),
            format_date(self.end),
            self.organization or "",
            self.title or "",
            serialize_dict(self.title_ids),
            self.item or "",
            serialize_dict(self.item_ids),
            self.item_publication_date or "",
            authors,
            serialize_dict(self.dimension_data),
            self.metric or "",
            str(self.value),
        )

    @classmethod
    def empty_generator(cls) -> typing.Generator["CounterRecord", None, None]:
        empty: typing.List["CounterRecord"] = []
        return (e for e in empty)



"""Celus format for non-counter data"""

import csv
import typing
import warnings
from datetime import date
from functools import lru_cache

from celus_nigiri import CounterRecord
from celus_nigiri.utils import parse_date_fuzzy

DEFAULT_COLUMN_MAP = {
    "Metric": "metric",
    "Organization": "organization",
    "Source": "title",
    "Title": "title",
}


@lru_cache
def col_name_to_month(row_name: str) -> typing.Optional[date]:
    """
    >>> col_name_to_month('Jan 2019')
    datetime.date(2019, 1, 1)
    >>> col_name_to_month('2019-02')
    datetime.date(2019, 2, 1)
    >>> col_name_to_month('prase') is None
    True
    """
    _date = parse_date_fuzzy(row_name)
    if _date:
        return _date.replace(day=1)
    return None


def custom_data_to_records(
    records: typing.Generator[dict, None, None], column_map=None, extra_dims=None, initial_data=None
) -> typing.Generator[CounterRecord, None, None]:
    warnings.warn(
        "This function is deprecated please use celus-nibbler to parse celus format",
        DeprecationWarning,
    )

    # prepare the keyword arguments
    if initial_data is None:
        initial_data = {}
    if column_map is None:
        column_map = DEFAULT_COLUMN_MAP
    if extra_dims is None:
        extra_dims = []
    # process the records
    result = []
    for record in records:
        implicit_dimensions = {}
        explicit_dimensions = {}
        monthly_values = {}
        for key, value in record.items():
            month = col_name_to_month(key)
            if month:
                monthly_values[month] = int(value)
            else:
                if key in column_map:
                    implicit_dimensions[column_map[key]] = value
                elif key in extra_dims:
                    explicit_dimensions[key] = value
                else:
                    raise KeyError(f'We don\'t know how to interpret the column "{key}"')
        # we put initial data into the data we read - these are usually dimensions that are fixed
        # for the whole import and are not part of the data itself
        for key, value in initial_data.items():
            if key not in implicit_dimensions:
                implicit_dimensions[key] = value  # only update if the value is not present
        for month, value in monthly_values.items():
            result.append(
                CounterRecord(
                    value=value,
                    start=month,
                    dimension_data=explicit_dimensions,
                    **implicit_dimensions,
                )
            )
    return (e for e in result)  # TODO convert this into a propper generator


def get_months(
    file: typing.IO,
    seek_fun: typing.Optional[typing.Callable[[csv.reader], None]] = None,
    dialect=None,
) -> typing.List[date]:
    """Get months from file in celus csv format"""
    pos = file.tell()
    reader = csv.reader(file, dialect=dialect)
    if seek_fun:
        seek_fun(reader)
    header = next(reader)

    res = []
    for cell in header:
        if date := col_name_to_month(cell.strip()):
            res.append(date)

    file.seek(pos)
    return res



"""
Module dealing with data in the COUNTER51 format.
"""

import json
import typing
from abc import ABCMeta, abstractmethod
from copy import deepcopy
from datetime import date, datetime

import ijson.backends.yajl2_c as ijson

from celus_nigiri.record import Author, CounterRecord

from .counter5 import ALLOWED_ITEM_IDS, CounterError, TransportError
from .error_codes import ErrorCode
from .exceptions import SushiException
from .utils import get_date_range, parse_date_fuzzy


class Counter51ReportBase(metaclass=ABCMeta):
    extra_params: typing.Dict[str, str] = {}

    @property
    @abstractmethod
    def dimensions(self) -> typing.List[str]:
        raise NotImplementedError()

    @property
    @abstractmethod
    def allowed_item_ids(self) -> typing.Dict[str, str]:
        raise NotImplementedError()

    def __init__(self, report: typing.Optional[typing.IO[bytes]] = None, http_status_code=None):
        self.records = []
        self.queued = False
        self.record_found: bool = False  # is populated once `fd_to_dicts` is called
        self.header = {}
        self.errors: typing.List[typing.Union[CounterError, TransportError]] = []
        self.warnings: typing.List[CounterError] = []
        self.infos: typing.List[CounterError] = []
        self.http_status_code = http_status_code

        # Parse it for the first time to extract errors, warnings and infos
        if report:
            self.fd_to_dicts(report)

    def read_report(
        self, header: dict, report_items: typing.Generator[dict, None, None]
    ) -> typing.Generator[CounterRecord, None, None]:
        """
        Reads in the report as returned by the API using Sushi51Client
        """
        for report_item in report_items:
            self.check_item(report_item)
            title = self._item_get_title(report_item)
            title_ids = self._item_get_title_ids(report_item)
            attribute_performances = report_item.get("Attribute_Performance", [])
            for ap in attribute_performances:
                dimension_data = self._extract_dimension_data(self.dimensions, report_item, ap)
                performances = ap.get("Performance", {})
                for metric, data in performances.items():
                    for month, value in data.items():
                        start = datetime.strptime(month, "%Y-%m").date().replace(day=1)

                        yield CounterRecord(
                            value=int(value),
                            metric=metric,
                            start=start,
                            title=title,
                            title_ids=title_ids,
                            dimension_data=dimension_data,
                        )

    def check_header(self, header, fd):
        lower_keys = [e.lower() for e in header]
        for field in ["release", "report_id", "report_name"]:  # mandatory header fields
            if field not in lower_keys:
                raise SushiException("Incorrect format", content=fd.read())

    def check_item(self, item: dict):
        """Check whether the item is valid"""
        self.check_item_root(item)
        if isinstance(item["Attribute_Performance"], list):
            for ap in item["Attribute_Performance"]:
                self.check_performance_extra_ids(ap)

    def check_item_root(self, item: dict):
        if "Attribute_Performance".lower() not in (e.lower() for e in item.keys()):
            raise SushiException(
                "Missing 'Attribute_Performance' field",
                content=json.dumps(item),
            )

    def check_performance_extra_ids(self, item: dict):
        """Check that there are no extra title IDs"""
        allowed = {e.lower() for e in self.allowed_item_ids}
        ids_keys = {e.lower() for e in (item.get("Item_ID", {}) or {}).keys()}
        extra = ids_keys - allowed
        if extra:
            raise SushiException(f"Extra IDs not allowed {extra}", content=json.dumps(item))

    def extract_errors(self, data):
        from .client import Sushi51Client

        sushi_errors = Sushi51Client.extract_errors_from_data(data)

        if any(str(e.code) == str(ErrorCode.PREPARING_DATA.value) for e in sushi_errors):
            # special status telling us we should retry later
            self.queued = True

        for sushi_error in sushi_errors:
            counter_error = CounterError.from_sushi_error(sushi_error)
            if sushi_error.is_warning:
                self.warnings.append(counter_error)
            elif sushi_error.is_info:
                self.infos.append(counter_error)
            else:
                self.errors.append(counter_error)

    def fd_to_dicts(
        self, fd: typing.IO[bytes]
    ) -> typing.Tuple[dict, typing.Generator[dict, None, None]]:
        def empty_generator() -> typing.Generator[dict, None, None]:
            empty: typing.List[dict] = []
            return (e for e in empty)

        # make sure that fd is at the beginning
        fd.seek(0)

        # first check whether it is not an error report
        # skip potential leading whitespace
        first_character = fd.read(1)  # type: bytes
        while first_character.isspace():
            first_character = fd.read(1)
        fd.seek(0)

        if first_character == b"[":
            # entire json can be a list of errors
            self.extract_errors(json.load(fd))
            return {}, empty_generator()

        elif first_character == b'"':
            # stringified header with an error received
            json_string = json.load(fd)
            data = json.loads(json_string)
            self.extract_errors(data)
            header = data.get("Report_Header", {})
            self.header = deepcopy(header)
            return data, empty_generator()

        elif first_character != b"{":
            self.errors.append(
                TransportError(http_status=self.http_status_code, message="Non SUSHI data returned")
            )
            return {}, empty_generator()

        # Try to read the data
        self.record_found = bool(next(ijson.items(fd, "Report_Items.item"), None))
        fd.seek(0)
        if self.record_found:
            items = ijson.items(fd, "Report_Items.item")
        else:
            self.record_found = bool(next(ijson.items(fd, "body.Report_Items.item"), None))
            if self.record_found:
                items = ijson.items(fd, "body.Report_Items.item")
            else:
                items = empty_generator()
        fd.seek(0)

        # try to read the header
        header = next(ijson.items(fd, "Report_Header"), None)
        fd.seek(0)
        # check whether the header is not located in 'body' element
        if not header:
            header = dict(ijson.kvitems(fd, "body.Report_Header"))
            fd.seek(0)

        if not self.record_found and not header:
            # not data and header not found entire json could be a header
            header = json.load(fd)
        fd.seek(0)

        # Extract errors
        if header:
            # Try to extract exceptions from header
            self.extract_errors(header)
            if not any((self.errors, self.warnings, self.infos)):
                # no  error/warning/info = > file has to contain a valid header
                self.check_header(header, fd)
                fd.seek(0)
            self.header = deepcopy(header)
        elif not self.record_found:
            # Header is missing and no data
            raise SushiException("Incorrect format", content=fd.read())
        else:
            # Header is empty, but data are present
            pass

        # error can be placed outside of header
        # look for them only if the header is missing or there are no data records
        # - if we look for them in each case, it can add quite some overhead for large files
        #   (like 20s for a 50 MB file)
        if (
            (not header or not self.record_found)
            and not self.errors
            and not self.warnings
            and not self.infos
        ):
            # Extract exceptions from root
            self.extract_errors(list(ijson.items(fd, "Exceptions.item")))  # In <root>
            fd.seek(0)
            self.extract_errors(dict(ijson.kvitems(fd, "Exception")))  # Single exception in <root>
            fd.seek(0)
            # Extract exception(s) from body
            self.extract_errors(list(ijson.items(fd, "body.Exceptions.item")))
            fd.seek(0)
            self.extract_errors(dict(ijson.kvitems(fd, "body.Exception")))
            fd.seek(0)

        return header, items

    def get_months(self, fd: typing.IO[bytes]) -> typing.List[date]:
        pos = fd.tell()

        header, _ = self.fd_to_dicts(fd)

        # Try to obtain End_Date and Begin_Date from report filter
        date_start_str = None
        date_end_str = None
        if report_filters := header.get("Report_Filters", {}):
            date_start_str = report_filters.get("Begin_Date")
            date_end_str = report_filters.get("End_Date")

        fd.seek(pos)

        date_start = parse_date_fuzzy(date_start_str or "")
        date_end = parse_date_fuzzy(date_end_str or "")
        if date_start and date_end:
            return get_date_range(date_start, date_end)
        else:
            return []

    def file_to_records(self, filename: str) -> typing.Generator[CounterRecord, None, None]:
        f = open(filename, "rb")  # file will be closed later (once generator struct is discarded)
        header, items = self.fd_to_dicts(f)
        return self.read_report(header, items)

    @classmethod
    def file_to_input(cls, filename: str):
        with open(filename, "r", encoding="utf-8") as infile:
            return json.load(infile)

    @classmethod
    def _item_get_title(cls, item):
        return item.get("Title")

    @classmethod
    def _item_get_title_ids(cls, item) -> typing.Dict[str, str]:
        return cls._extract_title_ids(item.get("Item_ID", {}) or {})

    @classmethod
    def _item_get_item_ids(cls, item) -> typing.Dict[str, str]:
        return {}

    @classmethod
    def _item_get_publication_date(cls, item) -> typing.Optional[str]:
        return None

    @classmethod
    def _item_get_authors(cls, item) -> typing.Optional[typing.List[Author]]:
        return None

    @classmethod
    def _item_get_item(self, item) -> typing.Optional[str]:
        return None

    @classmethod
    def _extract_title_ids(cls, values: dict) -> dict:
        ret = {}
        for name, value in values.items():
            if id_type := cls.allowed_item_ids.get(name):
                ret[id_type] = value
        return ret

    @classmethod
    def _extract_dimension_data(cls, dimensions: list, *parts: dict):
        ret = {}
        for dim in dimensions:
            for part in parts:
                if dim in part:
                    ret[dim] = part[dim]
                    break
        return ret


class Counter51DRReport(Counter51ReportBase):
    extra_params: typing.Dict[str, str] = {"attributes_to_show": "Access_Method"}
    dimensions = ["Access_Method", "Data_Type", "Publisher", "Platform"]
    allowed_item_ids = ALLOWED_ITEM_IDS["DR"]

    @classmethod
    def _item_get_title(cls, item):
        return item.get("Database")


class Counter51TRReport(Counter51ReportBase):
    extra_params: typing.Dict[str, str] = {
        "attributes_to_show": "YOP|Access_Method|Access_Type|Section_Type",
    }
    dimensions = [
        "Access_Type",
        "Access_Method",
        "Data_Type",
        "Section_Type",
        "YOP",
        "Publisher",
        "Platform",
    ]
    allowed_item_ids = ALLOWED_ITEM_IDS["TR"]


class Counter51PRReport(Counter51ReportBase):
    extra_params: typing.Dict[str, str] = {"attributes_to_show": "Access_Method"}
    dimensions = ["Access_Method", "Data_Type", "Platform"]
    allowed_item_ids = ALLOWED_ITEM_IDS["PR"]


class Counter51IRReport(Counter51ReportBase):
    extra_params: typing.Dict[str, str] = {
        "attributes_to_show": "Authors|Publication_Date|YOP|Access_Method"
        "|Access_Type|Article_Version",
        "include_parent_details": True,
    }
    dimensions = [
        "Access_Type",
        "Access_Method",
        "Data_Type",
        "YOP",
        "Publisher",
        "Platform",
        "Article_Version",
        "Parent_Data_Type",
    ]
    allowed_item_ids = ALLOWED_ITEM_IDS["IR"]

    def read_report(
        self, header: dict, report_items: typing.Generator[dict, None, None]
    ) -> typing.Generator[CounterRecord, None, None]:
        """
        Reads in the report as returned by the API using Sushi51Client
        """
        for report_item in report_items:
            title = self._item_get_title(report_item)
            title_ids = self._item_get_title_ids(report_item)
            # Data_Type can be present here for Title and
            # in nested data for Item we need to preserve the one for the title
            parent_data_type = report_item.get("Data_Type")
            for item in report_item.get("Items", []):
                self.check_item(item)
                item_name = self._item_get_item(item)
                item_ids = self._item_get_item_ids(item)
                item_publication_date = self._item_get_publication_date(item)
                item_authors = self._item_get_authors(item)
                attribute_performances = item.get("Attribute_Performance", [])
                for ap in attribute_performances:
                    dimension_data = self._extract_dimension_data(
                        self.dimensions, ap, item, report_item
                    )
                    if parent_data_type:
                        dimension_data["Parent_Data_Type"] = parent_data_type
                    performances = ap.get("Performance", {})
                    for metric, data in performances.items():
                        for month, value in data.items():
                            start = datetime.strptime(month, "%Y-%m").date().replace(day=1)

                            yield CounterRecord(
                                value=int(value),
                                metric=metric,
                                start=start,
                                title=title,
                                title_ids=title_ids,
                                item=item_name,
                                item_ids=item_ids,
                                item_publication_date=item_publication_date,
                                item_authors=item_authors,
                                dimension_data=dimension_data,
                            )

    def _item_get_item(self, data: dict):
        return data.get("Item", data)

    @classmethod
    def _item_get_item_ids(cls, item):
        return cls._extract_title_ids(item.get("Item_ID", []) or [])

    @classmethod
    def _item_get_publication_date(cls, item) -> typing.Optional[str]:
        return item.get("Publication_Date")

    @classmethod
    def _item_get_authors(cls, item) -> typing.Optional[typing.List[Author]]:
        res = []
        for author in item.get("Authors", []):
            res.append(
                Author(name=author.get("Name"), ISNI=author.get("ISNI"), ORCID=author.get("ORCID"))
            )
        return res or None



import itertools
import pathlib
import sys

from . import csv_line_length_histogram, detect_csv_dialect, detect_file_encoding


def main():
    root_path = pathlib.Path(sys.argv[1])

    paths = []
    if root_path.is_dir():
        paths = list(
            itertools.chain(
                root_path.glob("**/*.csv"),
                root_path.glob("**/*.tsv"),
            )
        )
    elif root_path.is_file():
        if root_path.suffix.lower() in [".tsv", ".csv"]:
            paths = [root_path]

    for path in paths:
        with path.open("rb") as f:
            encoding = detect_file_encoding(f)

        try:
            with path.open("r", encoding=encoding) as f:
                dialect = detect_csv_dialect(f)
                histogram = csv_line_length_histogram(f, dialect)
        except UnicodeDecodeError as e:
            print(f"{path} {e}")
        else:
            print(f"{path}: {dialect}|{encoding} - {histogram}")


if __name__ == "__main__":
    main()
